/*
 * msklog.c based on the OPenAFS cklog.c program with the
 * following copyright:
 */

/*
 * Copyright 2000, International Business Machines Corporation and others.
 * All Rights Reserved.
 *
 * This software has been released under the terms of the IBM Public
 * License.  For details, see the LICENSE file in the top-level source
 * directory or online at http://www.openafs.org/dl/license10.html
 */

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <time.h>
      
#include <des.h>
#include <afs/stds.h>
#include <afs/auth.h>
#include <afs/kautils.h>
#include <afs/ptserver.h>
#include <afs/cm_config.h>
#include <afs/cmd.h>
#include <afs/dirpath.h>


#include "msklog.h"

int CommandProc(struct cmd_syndesc *, char *);

static int debug = 0;
static int zero_argc;    /* so we can clear passwd on cmdline */
static char **zero_argv;

static char *getpipepass() {
    static char gpbuf[BUFSIZ];
    /* read a password from stdin, stop on \n or eof */
    register int i, tc;
    memset(gpbuf, 0, sizeof(gpbuf));
    for(i=0; i<(sizeof(gpbuf)-1); i++) {
    tc = fgetc(stdin);
    if (tc == '\n' || tc == EOF) break;
    gpbuf[i] = tc;
    }
    return gpbuf;
}

/* good_gets is like gets except that it take a max string length and won't
 * write past the end of its input buffer.  It returns a variety of negative
 * numbers in case of errors and zero if there was no characters read (a blank
 * line for instance).  Otherwise it returns the length of the string read in.
 */

static int good_gets (s, max)
  char *s;
  int   max;
{   int l;              /* length of string read */
    if (!fgets (s, max, stdin)) {
    if (feof(stdin)) return EOF;    /* EOF on input, nothing read */
    else return -2;         /* I don't think this can happen */
    }
    l = strlen (s);
    if (l && (s[l-1] == '\n')) s[--l] = 0;
    return l;
}

static int read_pw_string(char *s, int max)
{
    int ok = 0;
    HANDLE h;
    int md;

    /* set no echo */
    h = GetStdHandle (STD_INPUT_HANDLE);
    GetConsoleMode (h, &md);
    SetConsoleMode (h, md & ~ENABLE_ECHO_INPUT);

    while (!ok) {
    printf("Password:");
    fflush(stdout);
    if (good_gets(s, max) <= 0) {
        printf("\n"); fflush(stdout);
        if (feof (stdin)) break;    /* just give up */
        else continue;      /* try again: blank line */
    }
    ok = 1;
    }

    if (!ok)
    memset(s, 0, max);

    /* reset echo */
    SetConsoleMode (h, md);
    printf("\n"); fflush(stdout);

    s[max-1] = 0;           /* force termination */
    return !ok;
}

static
void hexdump(char * comment, void * in, int len)
{
    unsigned char *p = (unsigned char *)in;
    int i;

	fprintf(stderr,"%s",comment);
    for (i=0; i<len; i++)
    {
        if ((i & 3) == 0) fprintf(stderr, " ");
		if ((i & 31) == 0) fprintf(stderr,"\n    ");
        fprintf(stderr,"%02x", p[i]);
    }
    fprintf(stderr,"\n");
}


/******************************************************************/


int 
store_afs_token(
    struct ktc_principal *client,
	struct ktc_principal *server,
    afs_int32 tkt_type,
    unsigned char *ticket_p,
    int ticket_len,
    struct ktc_encryptionKey * session_key,
    afs_int32 starttime,
    afs_int32 endtime,
    int set_pag)
{
	int rc;
    struct ktc_token token;
	char username[BUFSIZ];  /* To hold client username structure */

	if (debug) {
    	fprintf(stderr,"afs_store_token\n");
	}

    token.startTime = starttime;
    token.endTime = endtime;
    memcpy((char *) &token.sessionKey, session_key, sizeof(token.sessionKey));
    token.kvno = (short )tkt_type;
    token.ticketLen = ticket_len;
    if (ticket_len > MAXKTCTICKETLEN) {
	fprintf(stderr, "ticket is too long (length %d). Maximum length accepted by AFS cache manager is %d\n", ticket_len, MAXKTCTICKETLEN);
	exit(1);
    }
    memcpy( (char *) token.ticket, (char*) ticket_p, ticket_len);

	/* 
	 * Use PTS to find ViceID for this user 
	 * Replace the actual principal with this number
	 */

	strcpy (username,client->name);
	if (client->instance[0])  
	{
		strcat (username, ".");
		strcat (username, client->instance);
    }
	if (debug) {
        fprintf(stderr,"username=%s\n",username);
	}

#ifndef _WIN32
	if (!pr_Initialize (0, AFSDIR_CLIENT_ETC_DIRPATH, server->cell))
	{
		status = pr_SNameToId (username, &viceId);
		pr_End();
	}

    if ((status == 0) && (viceId != ANONYMOUSID))
	{
        sprintf (client->name, "AFS ID %d", viceId);
    	strcpy(client->instance, "");
	}
#else
	/* On unix the pr_Initialize convers this to lower case!!!*/
	lcstring(server->cell, server->cell, sizeof(server->cell));

#endif

	lcstring(client->cell, client->cell, sizeof(client->cell));

	if (debug) {
		fprintf(stderr,"client->name=%s\n", client->name);
		fprintf(stderr,"client->instance=%s\n", client->instance);
		fprintf(stderr,"client->cell=%s\n",client->cell);
		fprintf(stderr,"server->name=%s\n", server->name);
		fprintf(stderr,"server->instance=%s\n", server->instance);
		fprintf(stderr,"server->cell=%s\n",server->cell);
		hexdump("token.sessionkey",&token.sessionKey, 
				sizeof(token.sessionKey));
		fprintf(stderr,"ticketsize=%04x\n",token.ticketLen);
		hexdump("TICKET", &token.ticket, token.ticketLen);
	}

    rc =  ktc_SetToken(server, &token, client,
			 set_pag ? AFS_SETTOK_SETPAG : 0);

	if (debug) {
        fprintf(stderr,"ktc_SetToken=%d\n",rc);
	}

	return rc;
}


/******************************************************************/

void
main(
    int argc,
    char *argv[]
    )
{
	int code;
	struct cmd_syndesc *ts;

	zero_argc = argc;
	zero_argv = argv;

	ts = cmd_CreateSyntax((char *) 0, CommandProc, 0, 
			"Obtain AFS token using Kerberos V5");

#define aPRINCIPAL 0
#define aPASSWORD  1
#define aCELL      2
#define aPIPE      3
#define aDEBUG     4

#ifdef FIX_KVNO
#define aFIXKVNO  5
#endif


	cmd_AddParm(ts, "-principal", CMD_SINGLE, CMD_OPTIONAL, "user name");
	cmd_AddParm(ts, "-password", CMD_SINGLE, CMD_OPTIONAL, "user's password");
	cmd_AddParm(ts, "-cell", CMD_SINGLE, CMD_OPTIONAL, "cell name");

	cmd_AddParm(ts, "-pipe", CMD_FLAG, CMD_OPTIONAL, "read password from stdin");
	cmd_AddParm(ts, "-debug", CMD_FLAG, CMD_OPTIONAL, "debug");

#ifdef FIX_KVNO
	cmd_AddParm(ts, "-fixkvno" ,CMD_SINGLE, CMD_OPTIONAL, "new kvno");
#endif
	
	code = cmd_Dispatch(argc, argv);
	exit(code);
}

/*****************************************************************/
int
CommandProc(struct cmd_syndesc *as,
			char *arock)
{
	int i;
	char * reason = NULL;
	char  defaultCell[256];
	char * username = NULL;
	char * password = NULL;
	char passwd[BUFSIZ];
	ULONG status;
	afs_int32 afskvno = RXKAD_TKT_TYPE_KERBEROS_V5;
	afs_uint32 startTime;
	afs_uint32 endTime;
	struct ktc_encryptionKey sessionKey;
	struct ktc_principal client;
	struct ktc_principal server;
	char * tkt = NULL;
	int tktlen = 0;
	char * pcname = NULL;
	char * pcrealm = NULL;

#if FIX_KVNO
	char * newtkt = NULL;
	int newtktlen = 0;
	int fixkvno = 1;
	int code = 0;
#endif

	/* blow away command line arguments */
	for (i=1; i<zero_argc; i++) {
		memset (zero_argv[i], 0, strlen(zero_argv[i]));
	}
	zero_argc = 0;

	/* needed for pioctl to work */
	afs_winsockInit();

	if (as->parms[aDEBUG].items) {
			debug++;
	}

	memset(&client, 0, sizeof(client));
	memset(&server, 0 , sizeof(server));

	strcpy(server.name,"afs");

	cm_GetRootCellName(defaultCell);

	if (as->parms[aCELL].items) {
		strncpy(server.cell, 
				as->parms[aCELL].items->data, 
				sizeof(server.cell));
	} else {
		strncpy(server.cell, defaultCell, sizeof(server.cell));
	}
	/* its afs/cell@realm where cell and realm are the same for now */
	strncpy(server.instance,server.cell,sizeof(server.instance));

	if (as->parms[aPASSWORD].items) {
		password = as->parms[aPASSWORD].items->data;
	}
	
	if (as->parms[aPRINCIPAL].items) {
		username = as->parms[aPRINCIPAL].items->data;
	}

	if (username && !password) {
		if (as->parms[aPIPE].items) {
			strncpy(passwd, getpipepass(), sizeof(passwd));
		}
		else {
			if (read_pw_string(passwd, sizeof(passwd)))
				reason = "can't read password from terminal";
			else if (strlen(passwd) == 0)
				reason = "zero length password is illegal";
			else
				reason = NULL;
			if (reason) {
				fprintf (stderr, "Unable to login because %s.\n", reason);
				exit(2);
			}
		}
		password = passwd;
	}

	status = msklog_get_ticket(username, /* null for LSA  */
					password,           /* null for LSA */
					server.cell,  /* need to get default cell */
					&tkt,
					&tktlen,
					&sessionKey,
					&startTime,
					&endTime,
					client.name,
					sizeof(client.name),
					client.cell,
					sizeof(client.cell));
	if (status) {
		LPVOID lpMsgBuf;
		FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | 
					  FORMAT_MESSAGE_FROM_SYSTEM |
					  FORMAT_MESSAGE_IGNORE_INSERTS,
					  NULL,
					  status,
					  MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
					  (LPTSTR) &lpMsgBuf,
					  0,
					  NULL);
		fprintf(stderr,"%s\n",lpMsgBuf);
//		MessageBox(NULL,(LPCTSTR)lpMsgBuf, "Error", MB_OK | MB_ICONINFORMATION);
		LocalFree(lpMsgBuf);
		goto cleanup;
	}

	if (debug) {
		fprintf(stderr,"Start=%d,end=%d %d\n",startTime, endTime, 
				endTime-startTime);
	}

#ifdef FIX_KVNO
	newtkt = tkt;
	newtktlen = tktlen;
	if (as->parms[aFIXKVNO].items) {
            fixkvno = atoi(as->parms[aFIXKVNO].items->data);

		code = fix_kvno(fixkvno,
					  tkt, 
                      tktlen,
					  &newtkt,
					  &newtktlen);

		if (code != 0) fprintf(stderr,"fix_kvno Failed %d\n",
					code);
	}
#endif
	
	if (store_afs_token( &client, 
						 &server, 
						 afskvno,
#ifdef FIX_KVNO
						 newtkt,
						 newtktlen,
#else
                         tkt,
						 tktlen,
#endif
						 &sessionKey,
						 startTime,
						 endTime,
						 0)) {
		printf("Failed\n");  
		exit(1);
	}
cleanup:
	return 0;
}
